@@ -5,13 +5,14 @@ from __future__ import division
 import argparse
 import sys
 import random
+import itertools
 
 import six
 import xxhash
 import spooky
 
 
-__version__ = '0.3.2'
+__version__ = '0.3.3'
 __all__ = [
     'sample_tuple', 'sample_line', 'reservoir',
     'main', 'parse_arguments',
@@ -28,7 +29,7 @@ def main(args=None, sin=sys.stdin, sout=sys.stdout):
         sep = a.sep
         rate = a.rate
         funcname = a.hash
-        seed = a.seed
+        seed = a.seed or 'DEFAULT_SEED'
 
         if col == -1:
             tuples = ((l,) for l in sin)
@@ -39,8 +40,9 @@ def main(args=None, sin=sys.stdin, sout=sys.stdout):
             write(l[-1])
     elif a.method == 'reservoir':
         size = int(a.rate)
-        seed = a.seed
-        for l in reservoir(sin, size, seed):
+        seed = a.seed or None
+        keep_order = a.order
+        for l in reservoir(sin, size, seed, keep_order):
             write(l)
 
 
@@ -51,13 +53,14 @@ def parse_arguments(args):
         help='sampling rate (in hash sampling mode) or reservior size (in reservior sampling mode)'
     )
     parser.add_argument(
-        '-s', '--seed', type=str, default='DEFAULT_SEED',
+        '-s', '--seed', type=str,
         help='seed for hash function (in hash sampling mode) or random seed (in reservior sampling mode)'
     )
     parser.add_argument('-c', '--col', type=int, default=-1, help='column index (starts from 0)')
     parser.add_argument('--method', type=str, default='hash', help='sampling method: hash (default) or reservoir')
     parser.add_argument('--hash', type=str, default='xxhash32', help='hash function: xxhash32 (default) or spooky32')
     parser.add_argument('--sep', type=str, default=',', help='column separator')
+    parser.add_argument('--order', action='store_true', help='preserve input order')
 
     argdict = parser.parse_args(args)
     argdict.sep = six.u(argdict.sep)
@@ -119,7 +122,32 @@ def sample_line(s, rate, funcname='xxhash32', seed='DEFAULT_SEED'):
     )
 
 
-def reservoir(s, size, seed=None):
+def partition_tuple(s, ratios, col, funcname='xxhash32', seed='DEFAULT_SEED'):
+    func = _hash_with_seed(funcname, seed)
+    dart_ticks = [0] + [sum(ratios[:i+1]) * 0xFFFFFFFF for i in range(len(ratios))]
+    dart_ticks[-1] = 0xFFFFFFFF
+
+    tees = itertools.tee(s, len(ratios))
+    ranges = [(dart_ticks[i], dart_ticks[i + 1]) for i in range(len(dart_ticks) - 1)]
+
+    def _create_generator(stream, low, high):
+        return (l for l in stream if low <= func(l[col]) < high)
+
+    return [
+        _create_generator(tee, low, high)
+        for tee, (low, high) in zip(tees, ranges)
+    ]
+
+
+def partition_line(s, ratios, funcname='xxhash32', seed='DEFAULT_SEED'):
+    tuples = ((l,) for l in s)
+    return (
+        (l[-1] for l in stream)
+        for stream in partition_tuple(tuples, ratios, 0, funcname, seed)
+    )
+
+
+def reservoir(s, size, seed=None, keep_order=False):
     """Perform reservoir sampling.
 
     >>> logs = (
@@ -137,6 +165,7 @@ def reservoir(s, size, seed=None):
     :param s: stream of anything
     :param size: sample size
     :param seed: optional seed (any hashable object)
+    :param keep_order: force elements in sample to respect input order
     :return: sampled list
     """
     if seed is not None:
@@ -146,18 +175,22 @@ def reservoir(s, size, seed=None):
     s = iter(s)
 
     # 1. Initial phase to fill reservoir
+    k = 0
     for i in range(size):
-        buckets.append(next(s))
+        buckets.append((k, next(s)))
+        k += 1
 
     # 2. Probabilistic update
-    k = size
     for l in s:
         position = random.randint(0, k)
         if position < size:
-            buckets[position] = l
+            buckets[position] = (k, l)
         k += 1
 
-    return buckets
+    if keep_order:
+        buckets = sorted(buckets, key=lambda x: x[0])
+
+    return [e[1] for e in buckets]
 
 
 def _hash_with_seed(funcname, seed):
